####
## Code for submitted manuscript "An approximate Bayesian computation approach to parameter estimation in a stochastic stage structured population model"
## Katherine Scranton, Jonas Knape, Perry de Valpine
## 10/7/2013
## katherine.scranton@yale.edu


SMCalg13<- function(orders, dataset, distList, thresholds, errH=1, M=10000, N=M, Tsteps=1, initBirthTimes=rep(0,10), maxInds=400, sampT=seq(0,10,by=2)){
	
	# Implementation of the SMC ABC algorithm for parameter estimation of a population model for a stage structured cohort 
	# dynamics are assumed to be density independent, with heterogeneous stage durations
	
	## dataset - y
	## the single dataset from which we are basing our inference
	# load("typThetas.R")
	# dataset <- repD[22,]
	# rm("repD")
	
	## distList - {pj}
	# the set of distance metrics p that calculate how similar a simulation is to the dataset
	# MUST BE GREATER THAN 1
	
	## thresholds 
	## gives the percentages to calculate the threshold schedule {es}
	
	## errH - h
	## gives the scale with which to tighten theta variance at each step
	
	# M = 10000
	# number of particles to simulate at first
	
	# N = 10000
	# batch size for resampling particles
	
	# Tsteps = 1
	# the number of steps in the filter
	
	# initBirth Times, maxInds, and sampT are inputs for the simulation model
	
	##############################
	
	priorScale <- c(10,10,1,10) 
	
	# set priors
	tmin <- 0*priorScale
	tmax <- 1*priorScale
	
	# set lists to save parameters, weights, and distances
	allThetas <-list()
	allDistances <- list()
	allWeights <- list()
	
	## At s=1
	# draw M theta from a prior of 4-dimensional theta
	allThetas[[1]] <- thetaPrior5(priorScale,n=M)
	ntheta <- length(priorScale)
	
	# simulate M datasets
	simDataSets <- matrix(0,nrow=M,ncol=18)	
	for (i in 1:M){			
		sam <- passParticletoC(allThetas[[1]][i,],initN=initBirthTimes,maxInd=maxInds)
		simDataSets[i,] <- as.vector(sam,mode="double")
	}
	
	# calculate distances
	nmetrics <- length(distList)
	allDistances[[1]] <- calcDistMetrics2(distList,dataset,simDataSets)
	
	# housekeeping
	rm(list=c("sam","simDataSets"))
	
	# assign weights
	allWeights[[1]] <- matrix(rep(1/M,M),nrow=M,ncol=1)
	
	#### for the rest of Tsteps iterations
	for (s in 2:(Tsteps+1)){	
		
		# set thresholds for step s
		es <- rep(0,nmetrics)
		for (j in 1:nmetrics){
			es[j] <- sort(allDistances[[s-1]][,j])[ M*thresholds[s-1] ]
		}
		
		# set perturbation kernal sd
		errSDs <- rep(0,ntheta) 
		for (i in 1:ntheta){
			errSDs[i] <- sqrt(errH*2*var(allThetas[[s-1]][,i]))
			
		}
		
		## Until we have M new particles
		allThetas[[s]] <- matrix(rep(0,ntheta),nrow=1,ncol=ntheta)
		allDistances[[s]] <- matrix(rep(0,nmetrics),nrow=1,ncol=nmetrics)
		
		countM <- 0
		while (countM < M){
			
			# draw N theta* from theta(s-1) with weights(s-1)
			propTheta <- sample(1:M,M,replace=TRUE,prob=allWeights[[s-1]][,1])
			
			# perturb thetai ~ N(theta*,sd)
			perTheta <- perturbTheta3(allThetas[[s-1]][propTheta,],errSDs)
			
			# exclude those prior(theta1)==0
			for (i in 1:ntheta){		
				ibool <- (perTheta[,i] < tmin[i])|(perTheta[,i] > tmax[i])
				perTheta <- perTheta[!ibool,]
			}
	
			# simulate from model xi ~ f(thetai)
			wN <- nrow(perTheta)
			simDataSets <- matrix(0,nrow=wN,ncol=18)
				
			for (i in 1:wN){		
				sam <- passParticletoC(perTheta[i,],initN=initBirthTimes,maxInd=maxInds)
				simDataSets[i,] <- as.vector(sam,mode="double")
			}
			
			# housekeeping
			rm(list=c("sam","propTheta","ibool"))	
			# calculate distances
			distances <- calcDistMetrics2(distList,dataset,simDataSets)
			
			# housekeeping
			rm("simDataSets")
			
			# Only retain particles that meet all thresholds
			eBool <- rep(TRUE,wN)
			for (j in 1:nmetrics){
				eBool <- eBool&(distances[,j] < es[j])
			}

			allThetas[[s]] <- rbind(allThetas[[s]],perTheta[eBool,])
			allDistances[[s]] <- rbind(allDistances[[s]],distances[eBool,])
			
			# housekeeping
			rm("eBool","perTheta","wN")
					
			# count number you have kept
			countM <- nrow(allThetas[[s]])-1
			print(countM)
			
		}
		
		# set these M as the new thetas and move to next iteration
		allThetas[[s]] <- allThetas[[s]][2:(M+1),]
		allDistances[[s]] <- allDistances[[s]][2:(M+1),]
				
		## set weights as in Beaumont et al 2009
		allWeights[[s]] <- matrix(1,nrow=M,ncol=1)
		
		for (i in 1:M){
			
			irep <- matrix(rep(as.numeric(allThetas[[s]][i,]),M),nrow=M,ncol=ntheta,byrow=TRUE)
			tdiff <- irep-allThetas[[s-1]]
			tdiff <- tdiff/matrix(rep(errSDs),nrow=M,ncol=ntheta,byrow=TRUE)
			
			phiprod <- matrix(1,nrow=M,ncol=1)
			for (k in 1:ntheta){
				 phiprod <- phiprod*dnorm(tdiff[,k])
			}	
			
			allWeights[[s]][i,] <- allWeights[[s]][i,]/sum(phiprod*allWeights[[s-1]])
		}
		
		#housekeeping
		rm("irep","tdiff","phiprod")
		allWeights[[s]] <- allWeights[[s]]/sum(allWeights[[s]])

		results <- list(allThetas,allDistances,allWeights)
		tname <- paste("SMCtempfile",floor(orders),".R",sep="")
		save(results,file=tname)
		proc.time()
	}

	# Left with N theta vectors
	results <- list(allThetas,allDistances,allWeights)
	
	results
}
	
calcDistMetrics2 <- function(distList,dataset,simDataSets){
	
	colnames(dataset) <- c("E1","E2","E3","E4","E5","E6","I1","I2","I3","I4","I5","I6","A1","A2","A3","A4","A5","A6")
	colnames(simDataSets) <- c("E1","E2","E3","E4","E5","E6","I1","I2","I3","I4","I5","I6","A1","A2","A3","A4","A5","A6")
	M <- nrow(simDataSets)
	distances <- matrix(0,nrow=M,ncol=length(distList))
	
	for (j in 1:length(distList)){			
		distances[,j] <- distList[[j]](dataset,simDataSets)
	}
	
	distances
}

perturbTheta3 <- function(thetaP, errSD){
		
	for (i in 1:4){
		thetaP[,i] <- thetaP[,i]+rnorm(nrow(thetaP),mean=0,sd=errSD[i])
	}
	
	thetaP
}

thetaPrior5 <- function(priorScale, n=1){
	
	thetai <- matrix(0,nrow=n,ncol=length(priorScale))
	for (i in 1:length(priorScale)){
		thetai[,i] <- priorScale[i]*runif(n,min=0,max=1)
	}
	
	thetai
}
